#include <math.h>
#include <unistd.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "conjunto.h"
#include "pilha.h"
#include "conv.h"

// Constante usada para comparacao de zero. 
//Para evitar erros de arredondamento.
const double Conjunto::SMALL = 1E-8;

// Construtor.
Conjunto::Conjunto(int num) {
  fecho = NULL;
  primeiro = 1;
  n = num;
  pontos = new Ponto[n + 1];
  index = new int[n];
  temp = new int[n];
  sleep_time = 100000;
}

// Destrutor.
Conjunto::~Conjunto() {
  delete [] pontos;
  delete [] index;
  delete [] temp;
}

// Imprime os pontos do conjunto.
void Conjunto::imprimePontos() {
  int i;

  glColor3f(1.0, 1.0, 1.0);
  glBegin(GL_POINTS);

  for (i = 0; i < n; i++)
    glVertex2f(pontos[i].x, pontos[i].y);

  glEnd();
}

// Insere um ponto no conjunto.
void Conjunto::inserePonto(int pos, double x, double y) {
  if ((pos < 0) || (pos >= n))
    return;
  if (primeiro) {
    primeiro = 0;
    esquerda = direita = x;
    emcima = embaixo = y;
  }
  if (esquerda > x)
    esquerda = x;
  if (direita < x)
    direita = x;
  if (emcima < y)
    emcima = y;
  if (embaixo > y)
    embaixo = y;

  pontos[pos].x = x;
  pontos[pos].y = y;
}

// Calcula o quadro da distancia entre dois pontos.
double Conjunto::Distancia_rt(Ponto &p1, Ponto &p2) {
  return (p1.x - p2.x) * (p1.x - p2.x) +
         (p1.y - p2.y) * (p1.y - p2.y);
}

// Calcula a area com sinal de um triangulo.
double Conjunto::Area2(Ponto &p1, Ponto &p2, Ponto &p3) {
  inicDesenha();
  glColor3f(1.0, 0.0, 0.0);
  glBegin(GL_LINE_LOOP);
  glVertex2f(p1.x, p1.y);
  glVertex2f(p2.x, p2.y);
  glVertex2f(p3.x, p3.y);
  glEnd();
  if (fecho != NULL)
    fecho->imprimeParcial();
  fimDesenha();
  usleep(sleep_time);

  global_cont++;

  return (ultimaArea = (p1.x * p2.y - p1.y * p2.x +
                        p1.y * p3.x - p1.x * p3.y +
                        p2.x * p3.y - p2.y * p3.x));
}

// Primitiva left. Diz se o terceiro ponto esta a esquerda dos 
// dois primeiros orientados do primeiro para o segundo.
bool Conjunto::Left(Ponto &p1, Ponto &p2, Ponto &p3) {
  return Area2(p1, p2, p3) > SMALL;
}

// Ordena os pontos em relacao a seus angulos polares com o
// primeiro ponto do conjunto.
void Conjunto::mergeSort(int *index, int inic, int fim) {
  int temp, temp2, i;
  int tam = fim - inic + 1;

  if (tam <= 1)
    return;
  if (tam == 2) {
    if (!Left(pontos[index[0]], pontos[index[inic]], pontos[index[fim]])) {
      temp = index[inic];
      index[inic] = index[fim];
      index[fim] = temp;
    }
    return;
  }

  int meio = inic + (tam / 2) - 1;

  mergeSort(index, inic, meio);
  mergeSort(index, meio + 1, fim);

  int *merge = new int[n];
  temp = i = inic;
  temp2 = meio + 1;
  while ((temp <= meio) && (temp2 <= fim))
    if (Left(pontos[index[0]], pontos[index[temp]], pontos[index[temp2]]))
      merge[i++] = index[temp++];
    else
      merge[i++] = index[temp2++];

  if (temp > meio)
    while (temp2 <= fim)
      merge[i++] = index[temp2++];
  else
    while (temp <= meio)
      merge[i++] = index[temp++];

  for (i = inic; i <= fim; i++)
    index[i] = merge[i];

  delete [] merge;
}

// Executa o algoritmo de Embrulho para Presente.
Poligono *Conjunto::Embrulho() {
  Poligono *fecho_atual;
  int i, min_y = 0;
  int ultimo, atual;

  fecho_atual = fecho = new Poligono();
  global_cont = 0;

  // Achar o ponto com menor y-coordenada.
  for (i = 1; i < n; i++)
    if (pontos[i].y < pontos[min_y].y)
      min_y = i;
    else
      if ((pontos[i].y == pontos[min_y].y) && (pontos[i].x > pontos[min_y].x))
        min_y = i;

  ultimo = min_y;

  // Procura as arestas do conjunto.
  // Quando ele acha pontos que sao colineares ele escolhe os pontos mais distantes.
  // Isso nao afeta a ordem do algoritmo pois o teste de distancia feito em O(1).
  do {
    fecho->insereVertex(pontos[ultimo]);
    if (ultimo == 0)
      atual = 1;
    else
      atual = 0;

    for (i = atual + 1; i < n; i++) {
      if (i == ultimo)
        continue;
      if (!(Left(pontos[ultimo], pontos[atual], pontos[i]))) {
        if ((fabs(ultimaArea) >= SMALL) || ((fabs(ultimaArea) < SMALL) &&
                                  (Distancia_rt(pontos[ultimo], pontos[atual]) <
                                   Distancia_rt(pontos[ultimo], pontos[i]))))
          atual = i;
      }
    }
    ultimo = atual;

  } while(ultimo != min_y);

  fecho = NULL;

  // Atualiza o contador de calculos de area.
  fecho_atual->setaCont(global_cont);
  return fecho_atual;
}

Poligono *Conjunto::Graham() {
  Poligono *fecho_atual;
  int i, j, min_y;
  int longe;
  Pilha p(n);

  fecho_atual = fecho = new Poligono();
  global_cont = 0;

  // Achar o ponto com menor y-coordenada.
  min_y = 0;
  for (i = 1; i < n; i++)
    if (pontos[i].y < pontos[min_y].y)
      min_y = i;
    else
      if ((pontos[i].y == pontos[min_y].y) && (pontos[i].x > pontos[min_y].x))
        min_y = i;

  // Fase de pre-processamento.
  index[0] = min_y;
  for(i = 0, j = 1; i < n; i++)
    if (i != min_y)
      index[j++] = i;

  mergeSort(index, 1, n - 1);

  // Coloca o ponto inicial na pilha.
  p.Empilha(index[0]);
  fecho->insereVertex(pontos[index[0]]);

  // No caso de existirem pontos colineares nessa parte ele seleciona
  //  o que estiver mais distante.
  // Os pontos que ficam para tras nunca mais serao olhados e portanto
  //  esses calculos, que sao da ordem de O(n), nao alteram a ordem do
  //  algoritmo.
  longe = 1;
  i = 2;
  while ((i < n) && (fabs(Area2(pontos[index[0]], pontos[index[longe]],
                                pontos[index[i]])) < SMALL)) {
    if (Distancia_rt(pontos[index[0]], pontos[index[longe]]) <
        Distancia_rt(pontos[index[0]], pontos[index[i]])) {
      longe = i;
    }
    i++;
  }
  if (i >= n)
    return fecho;
  p.Empilha(index[longe]);
  fecho->insereVertex(pontos[index[longe]]);

  // Insere, como o terceiro ponto na pilha, o primeiro ponto nao colinear
  //  com os anteriores.
  p.Empilha(index[i]);
  fecho->insereVertex(pontos[index[i]]);

  i++;
  
  // Calcula as arestas do fecho.
  while (i < n) {
    if (Left(pontos[p.Topo_1()], pontos[p.Topo()], pontos[index[i]])) {
      p.Empilha(index[i]);
      fecho->insereVertex(pontos[index[i]]);
      i++;
    }

    // Se o ponto nao estiver a esquerda verificamos se ele eh colinear
    //  ou se esta a direita, isso nao gasta outro calculo de area.
    else {
      if (fabs(ultimaArea) >= SMALL) {
        p.Desempilha();
        fecho->removeVertex();
      }

      // No caso do ponto ser colinear, pegamos o mais distante e partimos
      //  para o proximo ponto, como apenas trocamos o primeiro ponto da
      //  pilha, nao alteramos a propriedade que um ponto pode ser inserido
      //  mais de uma vez na pilha e como avancamos para o proximo ponto
      //  nao podemos entrar em loop. Portanto nao alteramos a ordem do
      //  algoritmo.
      else {
        if (Distancia_rt(pontos[p.Topo_1()], pontos[p.Topo()]) <
            Distancia_rt(pontos[p.Topo_1()], pontos[index[i]])) {
          p.Desempilha();
          fecho->removeVertex();
          p.Empilha(index[i]);
          fecho->insereVertex(pontos[index[i]]);
        }
        i++;
      }
    }
  }
  if (fabs(Area2(pontos[index[0]], pontos[p.Topo()],
           pontos[p.Topo_1()])) <= SMALL) {
    p.Desempilha();
    fecho->removeVertex();
  }

  fecho = NULL;

  fecho_atual->setaCont(global_cont);
  return fecho_atual;
}








