#include <stdio.h>
#include <stdlib.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "poligono.h"
#include "conjunto.h"
#include "conv.h"

const int LARGURA = 400;   // Dimensoes iniciais da tela
const int ALTURA = 300;

static Conjunto *S;              // Conjunto de pontos do arquivo de entrada.
static Poligono *conv = NULL;    // Fecho resultante
static FILE *in, *out;           // Arquivos de entrada e saida.
static int EMB = 1, GRA = 1;     // Booleanos para nao imprimir no arquivo de saida
                                 // mais de duas vezes

// Callback de display
void display() {
  inicDesenha();
  if (conv != NULL)
    conv->imprimePoligono();
  fimDesenha();
}

// Funcao que inicializa o desenho de uma nova tela.
void inicDesenha() {
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);
  S->imprimePontos();
}

// Funcao que desenha a nova tela.
void fimDesenha() {
  glutSwapBuffers();
}

// Callback de reshape
void reshape(int w, int h) {
  glViewport(0, 0, w, h);
}

// Callback de keyboard
void keyboard(unsigned char ch, int, int) {
  switch (ch) {

    // Sai do programa.
    case 'q' :
    case 'Q' : exit(0);

    // Limpa a tela.
    case 'c' :
    case 'C' : if (conv != NULL) {
                 delete conv;
                 conv = NULL;
               }
               glutPostRedisplay();
               break;

    // Executa o algoritmo de Graham.
    case 'g' :
    case 'G' : if (conv != NULL)
                 delete conv;
               conv = S->Graham();
	       printf("\nGraham-Scan:\n");
	       printf("Numero de AREAs: %d\n", conv->pegaCont());
	       if (GRA) {
		 GRA = 0;
		 fprintf(out, "\nGraham-Scan:\n");
		 if ((conv == 0) || ((conv != 0) && (conv->numVertex() <= 2)))
		   fprintf(out, "Nao existe fecho\n");
		 else {
		   conv->Imprime(out);
		   conv->imprimePoligono();
		 }
	       }
	       glutPostRedisplay();
               break;

    // Executa o algoritmo de Embrulho para Presente.
    case 'e' :
    case 'E' : if (conv != NULL)
                 delete conv;
               conv = S->Embrulho();
	       printf("\nEmbrulho para presente:\n");
	       printf("Numero de AREAs: %d\n", conv->pegaCont());
	       if (EMB) {
		 EMB = 0;
		 fprintf(out, "Embrulho para presente:\n");
		 if ((conv == 0) || ((conv != 0) && (conv->numVertex() <= 2)))
		   fprintf(out, "Nao existe fecho\n");
		 else {
		   conv->Imprime(out);
		   conv->imprimePoligono();
		 }
	       }
               glutPostRedisplay();
               break;
  }
}

// Imprime mensagem de erro de parametros de entrada.
void erroParam() {
  printf("Uso correto:\nconv <arq_entrada> <arq_saida> [<sleep_time in micro sec.>]\n\n");
}

// Le o arquivo de entrada.
Conjunto *leEntrada(FILE *in) {
  int n, i;

  // Le o numero de pontos.
  if (fscanf(in, "%d", &n) != 1) {
    printf("Erro na leitura do arquivo\n");
    exit(4);
  }

  // Cria um conjunto para colocar os pontos.
  Conjunto *novo = new Conjunto(n);

  float x, y;

  // Coloca os pontos no conjunto.
  for (i = 0; i < n; i++) {
    if (fscanf(in, "%f %f", &x, &y) != 2) {
      printf("Erro na leitura do arquivo\n");
      exit(4);
    }
    novo->inserePonto(i, x, y);
  }

  return novo;
}

int main(int argc, char *argv[]) {
  int par4;

  printf("1\n");

  // Verifica numero de parametros
  if ((argc != 3) && (argc != 4)) {
    erroParam();
    return 1;
  }

  printf("2\n");
  // Verifica a existencia do arquivo de entrada e abre o mesmo
  in = fopen(argv[1], "rt");
  if (in == NULL) {
    printf("Nao foi possivel abrir o arquivo de entrada \"%s\"\n", argv[1]);
    return 2;
  }

  printf("3\n");
  // Verifica se e possivel criar o arquivo de saida desejado.
  out = fopen(argv[2], "wt");
  if (out == NULL) {
    printf("Nao foi possivel abrir o arquivo de saida \"%s\"\n", argv[2]);
    return 3;
  }

  printf("4\n");
  // Le os pontos do arquivo de entrada.
  S = leEntrada(in);

  printf("5\n");
  // Verifica se foi dado um quarto parametro.
  if (argc == 4) {
    sscanf(argv[3], "%d", &par4);
    S->setSleepTime(par4);
  }
    
  printf("6\n");
  // Inicializacao do OpenGL e da janela.
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowSize(LARGURA, ALTURA);
  glutCreateWindow(argv[0]);

  printf("7\n");
  // Preparando Callbacks.
  glutReshapeFunc(reshape);
  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);

  printf("8\n");
  // Setando a projecao
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(S->Esquerda() - 1.0, S->Direita() + 1.0,
             S->Embaixo() - 1.0, S->Emcima() + 1.0);
  glMatrixMode(GL_MODELVIEW);

  conv = NULL;


  printf("9\n");
  // Loop principal do OpenGL, ele so retorna no final da execucao.
  glutMainLoop();

  // Fecha os arquivos
  fclose(in);
  fclose(out);

  // Libera a memoria usada.
  delete S;
  if (conv != NULL)
    delete conv;

  return 0;

}
